/* 
* Copyright (c) 2011 Yinxiao Li
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or (at
* your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "stdafx.h"
#include "Parallel_Skeletonization.h"
#include "blepo.h"
#include <list>

using namespace std;
using namespace blepo;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CWinApp theApp;

////----------local function---------/////

void FindSkeleton(ImgBinary* Input, list<CPoint>* tempskeleton)
{
	CPoint pt;
	for (pt.x=0;pt.x!=Input->Width()-1;pt.x++)
		for (pt.y=0;pt.y!=Input->Height()-1;pt.y++)		  
		{
			if ((*Input)(pt.x,pt.y)==1)
			{
				tempskeleton->push_back(pt);  // establish a frontier
			}
		}		
}


/*

Each time, the MyThin_FourDirections processes one direction of the current whole skeleton, 
given by parameter a, b, and c. The parameter tempskeleton returns the result of 
skeleton after one processing of direction .
If no more change in the size of the tempskeleton, the tempskeleton is the final skeleton.

*/

bool MyThin_FourDirections(ImgBinary* edges, int a, int b, int c, list<CPoint>& tempskeleton)
{  
	int i,j,index,sigma,chi;
	ImgBinary bimg,bimg1;
	int box[9];
    bool finish;

	finish=true;
	bimg=(*edges);
	bimg1=(*edges);
	
	list<CPoint>::iterator iter;

	for (iter=tempskeleton.begin();iter!=tempskeleton.end();)
	{ 
		sigma=0;
		index=0;

		for(j=-1;j!=2;j++)
			for (i=-1;i!=2;i++)			  
			{
				if ( (*iter).x+i>=0 && (*iter).x+i<bimg.Width()
					&& (*iter).y+j>=0 && (*iter).y+j<bimg.Height() )
				{
					box[index]=bimg((*iter).x+i,(*iter).y+j);
					index++;
				}
			}

			for (i=0;i<9;i++)
			{
				if(i==4) continue;
				sigma+=box[i];
			}
			
			chi=( box[3]!=box[1])+(box[1]!=box[5])+(box[5]!=box[7])+(box[7]!=box[3] )
				+2*( (!box[3]&&box[0]&&!box[1])+(!box[1]&&box[2]&&!box[5])+
				     (!box[5]&&box[8]&&!box[7])+(!box[7]&&box[6]&&!box[3]) );
			
			if ( (!box[a]&&box[b]&&box[c] ) && (chi==2) && (sigma!=1) )
			{
				bimg1((*iter).x,(*iter).y)=0;
				iter=tempskeleton.erase(iter);  // If satisfied, remove the point from the frontier.
				finish=false;
			}
			else iter++;
	}
	
	(*edges)=bimg1;	
	return finish;
	
}

//////-----------End of local functions--------------//////

/* 
Implements the Parallel Skeletonization algorithm, described in 
E. R. Davies and A. P. N. Plummer, Thinning Algorithm: a Critique and a New Methodology,
Pattern Recognition, 14(16), 1981.

*/

ImgBinary Skeleton(ImgBinary bimg)
{
    list<CPoint> tempskeleton;
	FindSkeleton(&bimg, &tempskeleton);
	
	while(1)
	{
		// If number of pixels of skeleton does not change any more, then stop.

		if(MyThin_FourDirections(&bimg,1,4,7,tempskeleton)) break;   // North	
		if(MyThin_FourDirections(&bimg,7,4,1,tempskeleton)) break;   // South
		if(MyThin_FourDirections(&bimg,3,4,5,tempskeleton)) break;   // West
		if(MyThin_FourDirections(&bimg,5,4,3,tempskeleton)) break;   // East		
	}	
	return(bimg);
}


//////-----------Start test main()--------------//////

int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
	int nRetCode = 0;
	
	// initialize MFC and print and error on failure
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
	{
		// TODO: change error code to suit your needs
		cerr << _T("Fatal Error: MFC initialization failed") << endl;
		nRetCode = 1;
	}
	else
	{
		ImgBgr img0;
		
		CString FilePathName;
		CFileDialog dlg(TRUE);
		if(dlg.DoModal()==IDOK)
			FilePathName=dlg.GetPathName(); 
		char *zStr = (char*)(LPCTSTR)FilePathName;
		
		for(int i=0;zStr[i]!='\0';i++)     
		{	if (zStr[i]=='\\')
		zStr[i]='/'; 
		}
		
		CString openfile = zStr;
		Load(openfile,&img0);
		
		ImgBinary img0_bi,img_skeleton;
		Convert(img0,&img0_bi);
		Figure test;
		test.Draw(img0_bi);
		
		clock_t start_time, end_time;

        start_time = clock();
		img_skeleton=Skeleton(img0_bi);
		end_time = clock();
		
		cout<< "time = "<<(1000 * (end_time - start_time)) / CLOCKS_PER_SEC <<" ms"<<endl;
		
		Figure skeleton;
		skeleton.Draw(img_skeleton);
		
	}
	EventLoop();
	return nRetCode;
}






